/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005-2006 EBM WebSourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: SourceUtilImpl.java 94 2006-03-26 17:11:05Z alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URI;
import java.nio.charset.Charset;

import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.commons.stream.ReaderInputStream;
import org.ow2.petals.commons.threadlocal.DocumentBuilders;
import org.ow2.petals.commons.threadlocal.Transformers;
import org.ow2.petals.component.framework.api.exception.DocumentException;
import org.ow2.petals.component.framework.api.exception.FaultException;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.SourceUtil;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Sample class used to manage the mapping between
 * <code>javax.xml.transform.Source</code> from String. the mapping is a quicly
 * "string" -> "<content>string</content>" conversion.
 * 
 * @author alouis - eBM WebSourcing
 * @author fgardes - eBM WebSourcing
 */
public class SourceUtilImpl implements SourceUtil {

    private final static String SOAPENV_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/";

    /**
     * Creates a new instance of {@link SourceUtilImpl}
     * 
     */
    public SourceUtilImpl() {
        super();
    }

    /**
     * Fork the given StreamSource
     * 
     * @param streamSource
     * @return
     */
    private static final InputStream forkStreamSource(final StreamSource streamSource) {
        InputStreamForker streamForker;
        final InputStream isContent = streamSource.getInputStream();
        if (isContent != null) {
            // The StreamSource was created from an InputStream
            streamForker = new InputStreamForker(isContent);
        } else {
            // The StreamSource was created from a Reader
            // we wrap it as an InputStream
            streamForker = new InputStreamForker(new ReaderInputStream(streamSource.getReader()));
        }
        streamSource.setInputStream(streamForker.getInputStreamOne());
        return streamForker.getInputStreamTwo();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createDocument
     * (javax.xml.transform.Source)
     */
    public Document createDocument(final Source source) throws PEtALSCDKException {
        return this.createDocument(source, true);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createDocument
     * (javax.xml.transform.Source, boolean)
     */
    public Document createDocument(final Source source, final boolean forkSource)
            throws PEtALSCDKException {
        Document document = null;
        try {
            if (source instanceof DOMSource) {
                final Node originalNode = ((DOMSource) source).getNode();
                if (originalNode instanceof Document) {
                    final Document originalDocument = (Document) originalNode;
                    if (forkSource) {
                        // clone the document
                        document = DocumentBuilders.getNamespaceDocumentBuilder().newDocument();
                        document.appendChild(document.importNode(originalDocument
                                .getDocumentElement(), true));
                    } else {
                        // refer the document
                        document = originalDocument;
                    }
                } else {
                    document = DocumentBuilders.getNamespaceDocumentBuilder().newDocument();
                    if (forkSource) {
                        // clone the node
                        document.appendChild(document.importNode(originalNode, true));
                    } else {
                        // adopt the node
                        document.appendChild(document.adoptNode(originalNode));
                    }

                }
            }

            if (document == null) {
                Source tempSource = source;
                if (forkSource && source instanceof StreamSource) {
                    tempSource = new StreamSource(forkStreamSource((StreamSource) source));
                }
                document = DocumentBuilders.getNamespaceDocumentBuilder().newDocument();
                final DOMResult domResult = new DOMResult(document);
                final Transformer transformer = Transformers.getDefaultTransformer();
                try {
                    transformer.transform(tempSource, domResult);
                } finally {
                    transformer.reset();
                }
            }
        } catch (final TransformerException e) {
            throw new PEtALSCDKException(e);
        }
        return document;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createSource(org
     * .w3c.dom.Document)
     */
    public DOMSource createDOMSource(final Document document) {
        // normalize the document to assure the resolution of each node
        // namespace
        document.normalizeDocument();
        return new DOMSource(document);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.util.SourceUtil#
     * createSoapFaultDocument(java.lang.Throwable, javax.xml.namespace.QName)
     */
    public Document createSoapFaultDocument(final Throwable e, final QName errorCode) {
        QName faultCode = errorCode;
        String faultString = e.getMessage();
        if (faultString == null) {
            faultString = "null message";
        }
        String details = "";
        URI faultActor = null;
        if (e instanceof DocumentException) {
            faultCode = ((DocumentException) e).getFaultCode();
            faultActor = ((DocumentException) e).getFaultActor();
            final Document document = ((DocumentException) e).getDocument();
            if (document != null && document.getDocumentElement() != null) {
                try {
                    details = XMLUtil.parseToString(document.getDocumentElement());
                } catch (final TransformerException e1) {
                    details = "An error occured when retrieving the exception document ("
                            + e1.getMessage() + ").";
                }
            }
        } else if (e instanceof FaultException) {
            faultCode = ((FaultException) e).getFaultCode();
            faultActor = ((FaultException) e).getFaultActor();
            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            details = sw.toString();

        } else {
            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            details = sw.toString();
        }
        // warning, the building of the document must be compatible DOM3 AND
        // DOM2
        final Document faultDocument = DocumentBuilders.getNamespaceDocumentBuilder().newDocument();

        final Element faultElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:Fault");
        // the NameSpace is not fill in an attribute with old XALAN DOM2
        // transformation
        if (!this.isDOM3(faultDocument)) {
            final Attr attr = faultDocument.createAttributeNS("http://www.w3.org/2000/xmlns/",
                    "xmlns:soapenv");
            attr.appendChild(faultDocument.createTextNode(SOAPENV_NAMESPACE));
            faultElement.setAttributeNode(attr);
        }
        faultDocument.appendChild(faultElement);

        final Element faultCodeElement = faultDocument.createElement("faultcode");
        if (faultCode != null) {
            if (faultCode.getNamespaceURI() != null) {
                if (faultCode.getNamespaceURI().equals(SOAPENV_NAMESPACE)) {
                    faultCodeElement.appendChild(faultDocument.createTextNode("soapenv:"
                            + faultCode.getLocalPart()));
                } else {
                    String faultCodePrefix = faultCode.getPrefix();
                    if (StringHelper.isNullOrEmpty(faultCodePrefix)) {
                        faultCodePrefix = "fcns";
                    }
                    final Attr attr = faultDocument.createAttributeNS(
                            "http://www.w3.org/2000/xmlns/", "xmlns:" + faultCodePrefix);
                    attr.appendChild(faultDocument.createTextNode(faultCode.getNamespaceURI()));
                    faultCodeElement.setAttributeNode(attr);
                    faultCodeElement.appendChild(faultDocument.createTextNode(faultCodePrefix + ":"
                            + faultCode.getLocalPart()));
                }
            } else {
                faultCodeElement
                        .appendChild(faultDocument.createTextNode(faultCode.getLocalPart()));
            }
        }
        // The node "faultstring" value must be enclosed in a CDATA section to
        // prevent encoding problems
        final Element faultStringElement = faultDocument.createElement("faultstring");
        faultStringElement.appendChild(faultDocument.createCDATASection(faultString));
        final Element faultDetailElement = faultDocument.createElement("detail");
        if (e instanceof DocumentException) {
            faultDetailElement.appendChild(faultDocument.createCDATASection(details));
        } else {
            final Element stackTrace = faultDocument.createElement("stackTrace");
            faultDetailElement.appendChild(stackTrace);
            stackTrace.appendChild(faultDocument.createCDATASection(details));
        }
        final Element faultActorElement = faultDocument.createElement("faultactor");
        if (faultActor != null) {
            faultActorElement.appendChild(faultDocument.createTextNode(faultActor.toString()));
        }
        faultElement.appendChild(faultCodeElement);
        faultElement.appendChild(faultStringElement);
        faultElement.appendChild(faultActorElement);
        faultElement.appendChild(faultDetailElement);
        return faultDocument;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createSource(java
     * .lang.String)
     */
    public Source createSource(final String message) throws PEtALSCDKException {
        final String charsetEncoding = Charset.defaultCharset().displayName();
        return this.createSource(message, charsetEncoding);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createSource(java
     * .lang.String, java.lang.String)
     */
    public Source createSource(final String message, final String charFormat)
            throws PEtALSCDKException {
        final StreamSource source = new StreamSource();
        try {
            final byte[] msgByte = message.getBytes(charFormat);
            final ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (final IOException e) {
            throw new PEtALSCDKException(e);
        }
        return source;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createStreamSource
     * (org.w3c.dom.Document)
     */
    public StreamSource createStreamSource(final Document document) throws PEtALSCDKException {

        final ByteArrayOutputStream bufferOut = new ByteArrayOutputStream();
        final Result sResult = new StreamResult(bufferOut);
        final Transformer transformer = Transformers.getDefaultTransformer();
        try {
            transformer.transform(new DOMSource(document), sResult);
        } catch (final TransformerException e) {
            throw new PEtALSCDKException("Error while transform DOM2 document to StreamSource", e);
        } finally {
            transformer.reset();
        }
        // TODO (CDE): Use CircularBuffer to increase performances.
        final ByteArrayInputStream bufferIn = new ByteArrayInputStream(bufferOut.toByteArray());
        return new StreamSource(bufferIn);

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createString(javax
     * .xml.transform.Source)
     */
    public String createString(final Source source) throws PEtALSCDKException {
        return this.createString(source, true);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#createString(javax
     * .xml.transform.Source, boolean)
     */
    public String createString(final Source source, final boolean forkSource)
            throws PEtALSCDKException {

        Source tempSource = source;
        if (forkSource && source instanceof StreamSource) {
            tempSource = new StreamSource(forkStreamSource((StreamSource) source));
        }

        try {
            final StringWriter buffer = new StringWriter();
            final Result sresult = new StreamResult(buffer);
            final Transformer transformer = Transformers.getDefaultTransformer();
            try {
                transformer.transform(tempSource, sresult);
            } finally {
                transformer.reset();
            }

            return buffer.toString();
        } catch (final TransformerConfigurationException e) {
            throw new PEtALSCDKException(e);
        } catch (final TransformerException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.util.SourceUtil#isDOM3(org.w3c
     * .dom.Document)
     */
    public boolean isDOM3(final Document document) {
        boolean result = true;

        try {
            document.getClass().getMethod("getBaseURI", (Class[]) null);
        } catch (final NoSuchMethodException e) {
            result = false;
        }

        return result;
    }

}
